/* -*- c-basic-offset: 3; -*- */
#ifndef DISKCACHE_TIME_HASH_HH
#define DISKCACHE_TIME_HASH_HH

#define DISKCACHE_API__TIME_HASH_DEBUG 1

// System Header Files   
#if DISKCACHE_API__TIME_HASH_DEBUG
extern "C" {
#include <pthread.h>
#endif /* DISKCACHE_API__TIME_HASH_DEBUG */
}

// System Header Files   
#include <utility>   
#include <map>   
#include <vector>   
#include <list>   
#include <utility>   
   
// General Header Files
#include "general/types.hh"
#include "general/ReadWriteLock.hh"
#include "general/SharedPtr.hh"

// Local Header Files   
#include "cachefile.hh"
#include "Query.hh"
   
#if DISKCACHE_API__TIME_HASH_DEBUG
#include "general/gpstime.hh"
#endif /* DISKCACHE_API__TIME_HASH_DEBUG */
   
namespace diskCache
{
   // Forward declarations
   class FrameQuery;
   
   
   //--------------------------------------------------------------------
   /// \brief Hash class to track available time ranges.
   /// 
   /// This class is designed to keep available time interval information
   /// for an IFO-Type pair.
   //--------------------------------------------------------------------
   class TimeHash
   {
   public:
   
      /// \brief Constructor
      TimeHash();
   
      /// \brief Destructor
      ~TimeHash();   

   
      /// \brief Structure to represent one interval's metadata
      //: (used as a value in time interval hash)
      struct IntervalInfo
      {
      public:
   
         /// \brief Constructors
         IntervalInfo();   
         IntervalInfo( const INT_4U stop_time, const INT_4U dt );
   
         /// \brief Writer
         void write( CacheFile& stream ) const;
   
         /// \brief Reader
         void read( CacheFile& stream );   

   
         /// \brief Stop time of the interval (GPS seconds)
         INT_4U mStopTime;
         
         /// \brief Time interval covered by a single file (in seconds)
         INT_4U mFileDt;

#if DISKCACHE_API__TIME_HASH_DEBUG
	 pthread_t		m_thread_id;
	 General::GPSTime	m_when;
#endif /* DISKCACHE_API__TIME_HASH_DEBUG */
      };
   
   
      /// \brief Known time hash under the same MOUNT_PT directory (resides under 
      //: different subdirectory). Used for data validation when inserting
      //: new frame data.
      struct KnownIntervals
      {
      public:
         typedef std::map< INT_4U, IntervalInfo > ranges_type;
           
	 //--------------------------------------------------------------
	 /// \brief Constructor   
	 //--------------------------------------------------------------
         KnownIntervals( );
        
	 //--------------------------------------------------------------
	 /// \brief Constructor   
	 //--------------------------------------------------------------
         KnownIntervals( const std::map< INT_4U, IntervalInfo > ranges );

         void setIntervals( const TimeHash* const ifo_hash );

         // FIX PR 2390: new query method: pass start,stop into search
         void setIntervals( INT_4U start, IntervalInfo ranges );

	 INT_4U CountFiles( ) const;

	 ranges_type& Ranges( );

	 const ranges_type& Ranges( ) const;

      private:

         // Allow data members access for TimeHash class.
         friend class TimeHash;
   
	 //--------------------------------------------------------------
         /// \brief Map of known time ranges
	 ///
	 /// StartTime <-> (StopTime,dT)
	 //--------------------------------------------------------------
         ranges_type mRanges;
      };
   
      //-----------------------------------------------------------------
      /// \brief Add an interval to the container.
      ///
      /// \param[in] GPSStart
      ///     The GPS start time in integral seconds.
      /// \param[in] DT
      ///     The lengt of the interval in seconds.
      //-----------------------------------------------------------------
      void Add( INT_4U GPSStart,
		INT_4U DT );

      void Add( const KnownIntervals& Source );

      size_t Count( ) const;

      /// \brief Test for emptiness of hash table.
      const bool empty() const;

      //-----------------------------------------------------------------
      /// \brief Check if the GPS time and delta time exists in hash
      ///
      /// \param[in] GPSStart
      ///     The GPS start time.
      /// \param[in] DT
      ///     The delta time.
      ///
      /// \return
      ///     If an entry bounded by GPSStart and DT exists, then
      ///     true is returned, false otherwise.
      //-----------------------------------------------------------------
      bool Find( INT_4U GPSStart, INT_4U DT ) const;

      void Find( const Query& SearchParams,
		 KnownIntervals& Intervals ) const;

      /// \brief Get hash info as a tcl list
      const std::string getTclHash( const std::string& record, 
                                    const INT_4S mtime ) const;

      /// \brief Writer
      void write( CacheFile& stream ) const;
   
      /// \brief Reader
      void read( CacheFile& stream );
   
      //-----------------------------------------------------------------
      /// \brief Determine if the bounding information is already in hash
      ///
      /// \param[in] GPSStart
      ///     The inclusive start time of the interval to search
      /// \param[in] GPSEnd
      ///     The exclusive end time of the interval to search
      ///
      /// \return
      ///     The value of true if any portion of
      ///     [GPSStart, GPSEnd) is part of the hash;
      ///     false otherwise.
      //-----------------------------------------------------------------
      bool Overlapping( INT_4U GPSStart, INT_4U GPSEnd ) const;

      /// \brief Usefull typedefs
      typedef std::map< INT_4U, IntervalInfo >::const_iterator MConstIter;   
      typedef std::pair< MConstIter, MConstIter > MConstIterPair;      
   
   
   private:
      //-----------------------------------------------------------------
      /// \brief Container of time intervals
      ///
      /// This data type indexes IntervalInfo based on GPS start times
      /// values.
      //-----------------------------------------------------------------
      typedef std::map< INT_4U, IntervalInfo > interval_container_type;
      typedef interval_container_type::iterator MIter;
      typedef std::pair< MIter, MIter > MIterPair;      
   
      typedef interval_container_type::value_type MValueType;      

      /// \brief Friends
      // To allow KnownIntervals to gain access to the mIntervals
      friend void KnownIntervals::setIntervals( const TimeHash* const ranges );   
   
   
      MIter insert( MIter Pos, const MValueType& Value );

      /// \brief Accessors
      const interval_container_type getIntervals() const;
   
   
      /// \brief Object lock
      mutable General::ReadWriteLock::lock_type mLock;
   
      /// \brief Map of StartTime<->{ StopTime, FileDt } time intervals
      interval_container_type mIntervals;   

   };
   
#if old
   inline const std::string& TimeHash::KnownIntervals::
   Path( ) const
   {
      return mDir;
   }
#endif /* old */
   
#if old
   inline const std::string&  TimeHash::KnownIntervals::
   IfoType( ) const
   {
      return mIfoType;
   }
#endif /* old */

   inline TimeHash::KnownIntervals::ranges_type& TimeHash::KnownIntervals::
   Ranges( )
   {
      return mRanges;
   }

   inline const TimeHash::KnownIntervals::ranges_type& TimeHash::KnownIntervals::
   Ranges( ) const
   {
      return mRanges;
   }

   inline CacheFile&
   operator<<( CacheFile& Stream,
	       TimeHash* const Data )
   {
      if ( Data )
      {
	 Data->write( Stream );
      }
      else
      {
	 TimeHash	data;

	 data.write( Stream );
      }
      return Stream;
   }
   
   inline CacheFile&
   operator<<( CacheFile& Stream,
	       General::SharedPtr< TimeHash > Data )
   {
      if ( Data.get( ) )
      {
	 Data->write( Stream );
      }
      else
      {
	 TimeHash	data;

	 data.write( Stream );
      }
      return Stream;
   }
   
   inline CacheFile&
   operator>>( CacheFile& Stream,
	       TimeHash*& Data )
   {
      Data = new TimeHash;

      Data->read( Stream );

      if ( Data->empty( ) )
      {
	 delete Data;
	 Data = (TimeHash*)NULL;
      }
      return Stream;
   }
   
   inline CacheFile&
   operator>>( CacheFile& Stream,
	       std::auto_ptr< TimeHash >& Data )
   {
      Data.reset( new TimeHash );

      Data->read( Stream );

      if ( Data->empty( ) )
      {
	 Data.reset( );
      }
      return Stream;
   }
   
}
      
   
#endif   
