#include "config.h"

#include <stdexcept>
#include <iostream>
#include <list>
#include <sstream>

#include "general/AtExit.hh"
#include "general/CommandLineOptions.hh"
#include "general/Thread.hh"
#include "general/unordered_map.hh"

#include "genericAPI/Logging.hh"
#include "genericAPI/Symbols.hh"

#include "cachefile.hh"
#include "diskcachecmd.hh"
#include "Query.hh"

using std::ostream;
using diskCache::Query;

//=======================================================================
// Global variables
//=======================================================================

static std::string	CacheFilename;

//=======================================================================
/// \brief Maintains a list of commands that are supported.
//=======================================================================
class CommandTable
{
public:
  enum command_type {
    CMD_UNKNOWN,
    CMD_DAEMON,
    CMD_DUMP,
    CMD_FILENAMES,
    CMD_FILENAMES_RDS,
    CMD_INTERVALS,
    CMD_MOUNT_POINT_STATS,
    CMD_QUIT,
    CMD_SCAN
  };

  //---------------------------------------------------------------------
  /// \brief Initialization of static data
  //---------------------------------------------------------------------
  static bool InitializeStatic( );

  //---------------------------------------------------------------------
  /// \brief Translate from ascii to enumerated type
  ///
  /// \param[in] Name
  ///     The ascii form of the command
  ///
  /// \return
  ///     The appropriate value from the command_type enumeration.
  ///     If Name cannot be mapped, then the value CMD_UNKNOWN is
  ///     returned.
  //---------------------------------------------------------------------
  static command_type Lookup( const std::string& Name );

private:
  typedef General::unordered_map< std::string, command_type >
  command_table_type;

  static bool	initialized;

  static command_table_type	command_table;
};

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
inline CommandTable::command_type CommandTable::
Lookup( const std::string& Name )
{
  command_table_type::const_iterator	pos( command_table.find( Name ) );

  //---------------------------------------------------------------------
  // Check if the command exists in the command table
  //---------------------------------------------------------------------
  if ( pos != command_table.end( ) )
  {
    //-------------------------------------------------------------------
    // It does; return the appropriate enumuration.
    //-------------------------------------------------------------------
    return pos->second;
  }
  //---------------------------------------------------------------------
  // It does not so return unknown command.
  //---------------------------------------------------------------------
  return CMD_UNKNOWN;
}

typedef General::CommandLineOptions 	CommandLineOptions;
typedef CommandLineOptions::Option 	Option;
typedef CommandLineOptions::OptionSet	OptionSet;

//=======================================================================
/// \brief Namespace for commands
//=======================================================================
namespace Command
{
  class Daemon
  {
  public:
    Daemon( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_CONCURRENCY,
      OPT_EXTENSIONS,
      OPT_LOG,
      OPT_MOUNT_POINTS
    };
    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    typedef std::vector< std::string >	extensions_type;
    typedef std::vector< std::string >	mount_points_type;

    INT_4U		m_concurrency;
    extensions_type	m_extensions;
    std::string		m_log;
    mount_points_type	m_mount_points;
  };

  class Dump
  {
  public:
    Dump( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_IFO,
      OPT_OUTPUT,
      OPT_TYPE,
      OPT_VERSION
    };
    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    std::string				m_ifo;
    diskCache::CacheFile::mode_type	m_mode;
    std::string				m_output;
    std::string				m_type;
    diskCache::CacheFile::version_type	m_version;
  };

  class Filenames
  {
  public:
    Filenames( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_IFO_TYPE_LIST,
      OPT_START_TIME,
      OPT_END_TIME
    };
    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    std::string				m_ifo_type_list;
    Query::gps_time_type		m_time_start;
    Query::gps_time_type		m_time_stop;
  };

  class FilenamesRDS
  {
  public:
    FilenamesRDS( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_IFO,
      OPT_TYPE,
      OPT_START_TIME,
      OPT_END_TIME,
      OPT_RESAMPLE
    };
    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    std::string				m_ifo;
    bool				m_resample;
    Query::gps_time_type		m_time_start;
    Query::gps_time_type		m_time_stop;
    std::string				m_type;
  };

  class Intervals
  {
  public:
    Intervals( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_IFO_TYPE_LIST,
      OPT_START_TIME,
      OPT_END_TIME
    };
    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    std::string				m_ifo_type_list;
    Query::gps_time_type		m_time_start;
    Query::gps_time_type		m_time_stop;
  };

  class MountPointStats
  {
  public:
    MountPointStats( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_IFO,
      OPT_TYPE
    };
    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    std::string				m_ifo;
    std::string				m_type;
  };

  class Scan
  {
  public:
    Scan( CommandLineOptions& Args );

    void operator()( );

    static const OptionSet& Options( );

  private:
    enum {
      OPT_CONCURRENCY,
      OPT_EXTENSIONS,
      OPT_MOUNT_POINTS,
      OPT_OUTPUT,
      OPT_TYPE,
      OPT_VERSION
    };

    typedef std::vector< std::string >	extensions_type;
    typedef std::vector< std::string >	mount_points_type;

    static OptionSet&	m_options;

    static OptionSet& init_options( );

    CommandLineOptions	m_args;

    extensions_type			m_extensions;
    diskCache::CacheFile::mode_type	m_mode;
    mount_points_type			m_mount_points;
    std::string				m_output;
    std::string				m_type;
    diskCache::CacheFile::version_type	m_version;
  };

}


//=======================================================================
//
//=======================================================================

void Usage( const CommandLineOptions::option_type& ProgramName,
	    const OptionSet& Opts );

//=======================================================================
// Main
//=======================================================================
int
main( int ArgC, char** ArgV )
{
  General::Thread::StackSizeDefault( 64 * 1024 );
  updateFileExtList( ".gwf" );
  //---------------------------------------------------------------------
  // Variables for logging
  //---------------------------------------------------------------------
  static char cwd_buffer[ 2048 ];
  getcwd( cwd_buffer, sizeof( cwd_buffer ) / sizeof( *cwd_buffer ) );
  std::string	cwd( cwd_buffer );

  GenericAPI::APINameSet( "diskcache" );
  GenericAPI::Symbols::LDAS_ARCHIVE_DIR::Set( cwd );
  GenericAPI::Symbols::LDAS_LOG_DIR::Set( cwd );

  try
  {
    //-------------------------------------------------------------------
    // Collect information about the command line options
    //-------------------------------------------------------------------
    enum common_option_types {
      OPT_COMMON_CACHE_FILE,
      OPT_COMMON_DEBUG_LEVEL,
      OPT_COMMON_HELP
    };

    CommandLineOptions	options( ArgC, ArgV );
    OptionSet		common_options;

    common_options.
      Synopsis( "[common_options]"
		" sub_command [command_options] [command_args]" );

    common_options.
      Summary( "This is a multi-purpose program that is intended"
	       " to provide much information about a cache of"
	       " file names." );

    common_options.
      Add( Option( OPT_COMMON_CACHE_FILE,
		   "cache-file",
		   Option::ARG_REQUIRED,
		   "Provide the file name of the cache file to use",
		   "filename" ) );
    common_options.
      Add( Option( OPT_COMMON_DEBUG_LEVEL,
		   "debug-level",
		   Option::ARG_REQUIRED,
		   "Specify the debugging level",
		   "level" ) );

    common_options.
      Add( Option( OPT_COMMON_HELP,
		   "help",
		   Option::ARG_NONE,
		   "Display this message" ) );

    common_options.Add( Command::Daemon::Options( ) );
    common_options.Add( Command::Dump::Options( ) );
    common_options.Add( Command::Filenames::Options( ) );
    common_options.Add( Command::FilenamesRDS::Options( ) );
    common_options.Add( Command::Intervals::Options( ) );
    common_options.Add( Command::MountPointStats::Options( ) );
    common_options.Add( Command::Scan::Options( ) );

    //-------------------------------------------------------------------
    // Parse the general arguments
    //-------------------------------------------------------------------
    try
    {
      std::string	arg_name;
      std::string	arg_value;
      bool		parsing = true;

      while( parsing )
      {
	int cmd_id
	  = options.Parse( common_options, arg_name, arg_value );

	switch( cmd_id )
	{
	case OPT_COMMON_CACHE_FILE:
	  CacheFilename = arg_value;
	  break;
	case OPT_COMMON_DEBUG_LEVEL:
	  {
	    std::istringstream	dbg_level_stream( arg_value );
	    int			dbg_level;

	    dbg_level_stream >> dbg_level;

	    GenericAPI::setLogDebugLevel( dbg_level );
	  }
	  break;
	case OPT_COMMON_HELP:
	  Usage( options.ProgramName( ), common_options );
	  General::AtExit::Cleanup( );
	  exit( 0 );
	  break;
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	}
      }
    }
    catch( ... )
    {
    }
    //-------------------------------------------------------------------
    // Get the command and evaluate it.
    //-------------------------------------------------------------------
    if ( options.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Read the cache file
      //-----------------------------------------------------------------
      if ( CacheFilename.empty( ) == false )
      {
	readDirCache( CacheFilename.c_str( ) );
      }

      //-----------------------------------------------------------------
      // There is another command. Get it and take the appropriate action
      //-----------------------------------------------------------------
      const CommandLineOptions::option_type cmd( options.Pop( ) );

      switch( CommandTable::Lookup( cmd ) )
      {
      case CommandTable::CMD_DAEMON:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::Daemon	cmd( options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_DUMP:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::Dump	cmd( options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_FILENAMES:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::Filenames	cmd( options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_FILENAMES_RDS:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::FilenamesRDS	cmd( options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_INTERVALS:
	{
	  //-------------------------------------------------------------
	  // Dump the time range information
	  //-------------------------------------------------------------
	  Command::Intervals	cmd( options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_MOUNT_POINT_STATS:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::MountPointStats	cmd( options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_SCAN:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::Scan	cmd( options );

	  cmd( );
	}
	break;
      default:
	{
	  std::ostringstream	msg;

	  msg << "Unknown command: " << cmd;

	  throw std::runtime_error( msg.str( ) );
	}
	break;
      }
    }
  }
  catch( const std::exception& Exception )
  {
    std::cerr << "FATAL: "
	      << Exception.what( )
	      << std::endl
      ;
    General::AtExit::Cleanup( );
    exit( 1 );
  }
  catch( ... )
  {
    std::cerr << "FATAL: "
	      << "unknown exception"
	      << std::endl
      ;
    General::AtExit::Cleanup( );
    exit( 1 );
  }
  //---------------------------------------------------------------------
  // Everything went as expected
  //---------------------------------------------------------------------
  General::AtExit::Cleanup( );
  return 0;
}

//=======================================================================
// CommandTable
//=======================================================================

CommandTable::command_table_type CommandTable::command_table;

bool CommandTable::initialized = CommandTable::InitializeStatic( );

bool CommandTable::
InitializeStatic( )
{
  static bool called = false;

  if ( called )
  {
    return called;
  }

  called = true;
  command_table[ "daemon" ] = CMD_DAEMON;
  command_table[ "dump" ] = CMD_DUMP;
  command_table[ "filenames" ] = CMD_FILENAMES;
  command_table[ "filenames-rds" ] = CMD_FILENAMES_RDS;
  command_table[ "intervals" ] = CMD_INTERVALS;
  command_table[ "mount-point-stats" ] = CMD_MOUNT_POINT_STATS;
  command_table[ "quit" ] = CMD_QUIT;
  command_table[ "scan" ] = CMD_SCAN;
  return called;
}

namespace Command
{
  //=====================================================================
  // Daemon
  //=====================================================================
  OptionSet& Daemon::m_options( Daemon::init_options( ) );

  OptionSet& Daemon::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: daemon" );

    retval.
      Summary( "The daemon sub command is intended to"
	       " provide a continuous scanning mode." );

    retval.Add( Option( OPT_CONCURRENCY,
			"concurrency",
			Option::ARG_REQUIRED,
			"Number of mount points to scan concurrently",
			"integer" ) );

    retval.Add( Option( OPT_EXTENSIONS,
			"extensions",
			Option::ARG_REQUIRED,
			"Comma seperated list of file extensions",
			"list" ) );

    retval.Add( Option( OPT_LOG,
			"log",
			Option::ARG_REQUIRED,
			"Specify where the log messages should be written.",
			"filename" ) );

    retval.Add( Option( OPT_MOUNT_POINTS,
			"mount-points",
			Option::ARG_REQUIRED,
			"Comma seperated list of mount points to scan",
			"list" ) );

    return retval;
  }

  Daemon::
  Daemon( CommandLineOptions& Args )
    : m_args( Args ),
      m_concurrency( 4 )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );

      while( parsing )
      {
	switch( m_args.Parse( m_options, arg_name, arg_value ) )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_CONCURRENCY:
	  {
	    std::istringstream	value_stream( arg_value );

	    value_stream >> m_concurrency;
	  }
	  break;
	case OPT_EXTENSIONS:
	  {
	    //-----------------------------------------------------------
	    // Generate list of extensions
	    //-----------------------------------------------------------
	    size_t pos = 0;
	    size_t end = 0;

	    while ( end != std::string::npos )
	    {
	      end = arg_value.find_first_of( ",", pos );
	      m_extensions.push_back( arg_value.substr( pos,( ( end == std::string::npos )
							      ? end
							      : end - pos ) ) );
	      pos = end + 1;
	    }
	  }
	  break;
	case OPT_LOG:
	  m_log = arg_value;
	  break;
	case OPT_MOUNT_POINTS:
	  {
	    //-----------------------------------------------------------
	    // Generate list of mount points
	    //-----------------------------------------------------------
	    size_t pos = 0;
	    size_t end = 0;

	    while ( end != std::string::npos )
	    {
	      end = arg_value.find_first_of( ",", pos );
	      m_mount_points.push_back( arg_value.substr( pos,( ( end == std::string::npos )
								? end
								: end - pos ) ) );
	      pos = end + 1;
	    }
	  }
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& Daemon::
  Options( )
  {
    return m_options;
  }

  void Daemon::
  operator()( )
  {
    bool		first( true );
    std::ostringstream	data;

    //-------------------------------------------------------------------
    // Setup the logging stream
    //-------------------------------------------------------------------
    std::ofstream	log_stream_base;
    std::ostream*	log_stream = &std::cout;

    if ( m_log.empty( ) == false )
    {
      log_stream_base.open( m_log.c_str( ) );
      if ( log_stream_base.is_open( ) )
      {
	log_stream = &log_stream_base;
      }
    }

    //-------------------------------------------------------------------
    // Add the extensions
    //-------------------------------------------------------------------
    first = true;
    data.str( "" );

    for ( extensions_type::const_iterator
	    cur = m_extensions.begin( ),
	    last = m_extensions.end( );
	  cur != last;
	  ++cur )
    {
      if ( first )
      {
	first = false;
      }
      else
      {
	data << " ";
      }
      data << *cur;
    }
    updateFileExtList( data.str( ).c_str( ) );

    //-------------------------------------------------------------------
    // Add the mount points
    //-------------------------------------------------------------------
    first = true;
    data.str( "" );

    for ( mount_points_type::const_iterator
	    cur = m_mount_points.begin( ),
	    last = m_mount_points.end( );
	  cur != last;
	  ++cur )
    {
      if ( first )
      {
	first = false;
      }
      else
      {
	data << " ";
      }
      data << *cur;
    }
    updateMountPtList( data.str( ).c_str( ), false );

    //-------------------------------------------------------------------
    // Setup scanning parameters
    //-------------------------------------------------------------------
    ScanConcurrency( m_concurrency );

    //-------------------------------------------------------------------
    // Scan the directories
    //-------------------------------------------------------------------

    ScanMountPointListContinuously( log_stream );

    //-------------------------------------------------------------------
    // Update the cache files
    //-------------------------------------------------------------------
    DumpCacheDaemonStart( log_stream );
    //-------------------------------------------------------------------
    // Read commands to be executed.
    //-------------------------------------------------------------------
    char	line_buffer[ 2048 ];
    bool	quit = false;

    while ( ( quit == false )
	    && ( std::cin.getline( line_buffer, sizeof( line_buffer ), '\n' ) ) )
    {
      //-----------------------------------------------------------------
      // Create list of options.
      //-----------------------------------------------------------------
      char*	lb_opts[ 20 ];
      int	cur_opt = 0;
      char*	lb_pos = line_buffer;

      lb_opts[ cur_opt ] = lb_pos;
      while( *lb_pos )
      {
	if ( *lb_pos == ' ' )
	{
	  *lb_pos = '\0';
	  if ( ++lb_pos )
	  {
	    lb_opts[ ++cur_opt ] = lb_pos;
	  }
	  continue;
	}
	++lb_pos;
      }
      lb_opts[ ++cur_opt ] = (char*)NULL;
      
      //-----------------------------------------------------------------
      // Switch on the command specified by the user
      //-----------------------------------------------------------------
      CommandLineOptions			line_options( cur_opt, lb_opts );

      switch( CommandTable::Lookup( line_options.ProgramName( ) ) )
      {
      case CommandTable::CMD_QUIT:
	{
	  quit = true;
	}
	break;
      case CommandTable::CMD_DUMP:
	{
	  //-------------------------------------------------------------
	  // Dump the current contents of the cache
	  //-------------------------------------------------------------
	  Command::Dump	cmd( line_options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_FILENAMES:
	{
	  Command::Filenames	cmd( line_options );

	  cmd( );
	}
	break;
      case CommandTable::CMD_FILENAMES_RDS:
	{
	  Command::FilenamesRDS	cmd( line_options );

	  cmd( );
	}
	break;
      default:
	{
	  std::cerr << "Unable to process: " << lb_opts[ 0 ]
		    << std::endl
	    ;
	}
	break;
      }
    }
    //-------------------------------------------------------------------
    // Get read to terminate
    //-------------------------------------------------------------------
    if ( log_stream_base.is_open( ) )
    {
      log_stream_base.close( );
    }
  }

  //=====================================================================
  // Dump
  //=====================================================================
  OptionSet& Dump::m_options( Dump::init_options( ) );

  OptionSet& Dump::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: dump" );

    retval.
      Summary( "The dump sub command is intended to dump the contents"
	       " of the memory cache."
	       " The quantity of data returned may be limited by the"
	       " use of the following options." );

    retval.Add( Option( OPT_IFO,
			"ifo",
			Option::ARG_REQUIRED,
			"IFO pattern to use for search. (Default: all)",
			"pattern" ) );

    retval.Add( Option( OPT_OUTPUT,
			"output",
			Option::ARG_REQUIRED,
			"Filename for the output",
			"filename" ) );

    retval.Add( Option( OPT_TYPE,
			"type",
			Option::ARG_REQUIRED,
			"Type pattern to use for search. (Default: all)",
			"pattern" ) );

    retval.Add( Option( OPT_VERSION,
			"version",
			Option::ARG_REQUIRED,
			"Version of the binary diskcache dump format to output",
			"version" ) );

    return retval;
  }

  Dump::
  Dump( CommandLineOptions& Args )
    : m_args( Args ),
      m_ifo( "all" ),
      m_mode( diskCache::CacheFile::CF_MODE_BINARY ),
      m_output( "" ),
      m_type( "all" ),
      m_version( diskCache::CacheFile::CURRENT_VERSION )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );

      while( parsing )
      {
	switch( m_args.Parse( m_options, arg_name, arg_value ) )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_IFO:
	  m_ifo = arg_value;
	  break;
	case OPT_TYPE:
	  m_type = arg_value;
	  break;
	case OPT_OUTPUT:
	  m_output = arg_value;
	  break;
	case OPT_VERSION:
	  {
	    std::istringstream	version_stream( arg_value );

	    version_stream >> m_version;
	    diskCache::CacheFile::ValidateVersion( m_version );
	  }
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& Dump::
  Options( )
  {
    return m_options;
  }

  void Dump::
  operator()( )
  {
    if ( m_output.empty( ) )
    {
      std::string
	cache( getDirCache( m_ifo.c_str( ), m_type.c_str( ) ) );

      size_t	b = 0;
      size_t	e = cache.find( " /" );
      while ( 1 )
      {
	std::cout << cache.substr( b,
				   ( e == std::string::npos )
				   ? e
				   : ( e - b ) )
		  << std::endl
	  ;
	if ( e == std::string::npos )
	{
	  break;
	}
	b = e + 1;
	e = cache.find( " /", b );
      }
    }
    else
    {
      //-----------------------------------------------------------------
      /// \todo
      ///     When the output is binary, the query options should allow
      ///     the user to request a subset of the entire cache.
      //-----------------------------------------------------------------
      diskCache::CacheFile	cf;

      cf.writeToFile( m_output.c_str( ),
		      m_mode,
		      m_version );
    }
  }

  //=====================================================================
  // Filenames
  //=====================================================================
  OptionSet& Filenames::m_options( Filenames::init_options( ) );

  OptionSet& Filenames::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: filenames" );

    retval.
      Summary( "The filenames sub command is intended to"
	       " query the memory cache."
	       " Several search options are available to restrict"
	       " the set of filename returned."
	       );

    retval.Add( Option( OPT_IFO_TYPE_LIST,
			"ifo-type-list",
			Option::ARG_REQUIRED,
			"comma seperated list of <ifo>-<type>- entries to search",
			"<ifo>-<type>- list" ) );

    retval.Add( Option( OPT_START_TIME,
			"start-time",
			Option::ARG_REQUIRED,
			"GPS start time of interest",
			"gps_time" ) );

    retval.Add( Option( OPT_END_TIME,
			"end-time",
			Option::ARG_REQUIRED,
			"GPS end time of interest",
			"gps_time" ) );

    return retval;
  }

  Filenames::
  Filenames( CommandLineOptions& Args )
    : m_args( Args ),
      m_ifo_type_list( "" ),
      m_time_start( 0 ),
      m_time_stop( ~0 )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );
      int		opt;

      while( parsing )
      {
	opt = m_args.Parse( m_options, arg_name, arg_value );

	switch( opt )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_IFO_TYPE_LIST:
	  {
	    m_ifo_type_list = arg_value;

	    //-----------------------------------------------------------
	    // Replace the commas with spaces
	    //-----------------------------------------------------------
	    size_t pos = 0;

	    while( ( pos = m_ifo_type_list.find_first_of( ",", pos ) )
		   != std::string::npos )
	    {
	      m_ifo_type_list[ pos ] = ' ';
	    }
	  }
	  break;
	case OPT_START_TIME:
	  {
	    std::istringstream( arg_value ) >> m_time_start;
	  }
	  break;
	case OPT_END_TIME:
	  {
	    std::istringstream( arg_value ) >> m_time_stop;
	  }
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& Filenames::
  Options( )
  {
    return m_options;
  }

  void Filenames::
  operator()( )
  {
    std::string
      cache( getFileNames( m_ifo_type_list.c_str( ),
			   m_time_start,
			   m_time_stop ) );

    size_t	b = 0;
    size_t	e = cache.find( " /" );
    while ( 1 )
    {
      std::cout << cache.substr( b,
				 ( e == std::string::npos )
				 ? e
				 : ( e - b ) )
		<< std::endl
	  ;
      if ( e == std::string::npos )
      {
	break;
      }
      b = e + 1;
      e = cache.find( " /", b );
    }
  }

  //=====================================================================
  // FilenamesRDS
  //=====================================================================
  OptionSet& FilenamesRDS::m_options( FilenamesRDS::init_options( ) );

  OptionSet& FilenamesRDS::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: filenames-rds" );

    retval.
      Summary( "The filenames-rds sub command is intended to"
	       " query the memory cache for filenames that can be used for createRDS command."
	       " Several search options are available to restrict"
	       " the set of filename returned."
	       );

    retval.Add( Option( OPT_IFO,
			"ifo",
			Option::ARG_REQUIRED,
			"IFO",
			"ifo" ) );

    retval.Add( Option( OPT_END_TIME,
			"end-time",
			Option::ARG_REQUIRED,
			"GPS end time of interest",
			"gps_time" ) );

    retval.Add( Option( OPT_RESAMPLE,
			"resample",
			Option::ARG_NONE,
			"Specify that resampling will be done" ) );

    retval.Add( Option( OPT_START_TIME,
			"start-time",
			Option::ARG_REQUIRED,
			"GPS start time of interest",
			"gps_time" ) );

    retval.Add( Option( OPT_TYPE,
			"type",
			Option::ARG_REQUIRED,
			"Type specifier of the the filename.",
			"type" ) );

    return retval;
  }

  FilenamesRDS::
  FilenamesRDS( CommandLineOptions& Args )
    : m_args( Args ),
      m_ifo( "all" ),
      m_resample( false ),
      m_time_start( 0 ),
      m_time_stop( ~0 ),
      m_type( "all" )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );
      int		opt;

      while( parsing )
      {
	opt = m_args.Parse( m_options, arg_name, arg_value );

	switch( opt )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_IFO:
	  m_ifo = arg_value;
	  break;
	case OPT_TYPE:
	  m_type = arg_value;
	  break;
	case OPT_START_TIME:
	  {
	    std::istringstream( arg_value ) >> m_time_start;
	  }
	  break;
	case OPT_END_TIME:
	  {
	    std::istringstream( arg_value ) >> m_time_stop;
	  }
	  break;
	case OPT_RESAMPLE:
	  m_resample = true;
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& FilenamesRDS::
  Options( )
  {
    return m_options;
  }

  void FilenamesRDS::
  operator()( )
  {
    std::string
      cache( getRDSFrameFiles( m_ifo.c_str( ),
			       m_type.c_str( ),
			       m_time_start,
			       m_time_stop,
			       m_resample ) );

    size_t	b = 0;
    size_t	e = cache.find( " /" );
    while ( 1 )
    {
      std::cout << cache.substr( b,
				 ( e == std::string::npos )
				 ? e
				 : ( e - b ) )
		<< std::endl
	  ;
      if ( e == std::string::npos )
      {
	break;
      }
      b = e + 1;
      e = cache.find( " /", b );
    }
  }

  //=====================================================================
  // Intervals
  //=====================================================================
  OptionSet& Intervals::m_options( Intervals::init_options( ) );

  OptionSet& Intervals::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: intervals" );

    retval.
      Summary( "The intervals sub command is intended to"
	       " query the memory cache and return time intervals"
	       " of file data."
	       );

    retval.Add( Option( OPT_IFO_TYPE_LIST,
			"ifo-type-list",
			Option::ARG_REQUIRED,
			"comma seperated list of <ifo>-<type>- entries to search",
			"<ifo>-<type>- list" ) );

    retval.Add( Option( OPT_START_TIME,
			"start-time",
			Option::ARG_REQUIRED,
			"GPS start time of interest",
			"gps_time" ) );

    retval.Add( Option( OPT_END_TIME,
			"end-time",
			Option::ARG_REQUIRED,
			"GPS end time of interest",
			"gps_time" ) );

    return retval;
  }

  Intervals::
  Intervals( CommandLineOptions& Args )
    : m_args( Args ),
      m_ifo_type_list( "" ),
      m_time_start( 0 ),
      m_time_stop( ~0 )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );
      int		opt;

      while( parsing )
      {
	opt = m_args.Parse( m_options, arg_name, arg_value );

	switch( opt )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_IFO_TYPE_LIST:
	  {
	    m_ifo_type_list = arg_value;

	    //-----------------------------------------------------------
	    // Replace the commas with spaces
	    //-----------------------------------------------------------
	    size_t pos = 0;

	    while( ( pos = m_ifo_type_list.find_first_of( ",", pos ) )
		   != std::string::npos )
	    {
	      m_ifo_type_list[ pos ] = ' ';
	    }
	  }
	  break;
	case OPT_START_TIME:
	  {
	    std::istringstream( arg_value ) >> m_time_start;
	  }
	  break;
	case OPT_END_TIME:
	  {
	    std::istringstream( arg_value ) >> m_time_stop;
	  }
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& Intervals::
  Options( )
  {
    return m_options;
  }

  void Intervals::
  operator()( )
  {
    std::string
      cache( getIntervalsList( m_ifo_type_list.c_str( ),
			       m_time_start,
			       m_time_stop ) );

    size_t	b = 0;
    size_t	e = cache.find( " /" );
    while ( 1 )
    {
      std::cout << cache.substr( b,
				 ( e == std::string::npos )
				 ? e
				 : ( e - b ) )
		<< std::endl
	  ;
      if ( e == std::string::npos )
      {
	break;
      }
      b = e + 1;
      e = cache.find( " /", b );
    }
  }

  //=====================================================================
  // MountPointStats
  //=====================================================================
  OptionSet& MountPointStats::m_options( MountPointStats::init_options( ) );

  OptionSet& MountPointStats::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: mountPointStats" );

    retval.
      Summary( "The mount-point-stats sub command is intended to get some statistical information"
	       " the memory cache grouped by the mount points."
	       );

    retval.Add( Option( OPT_IFO,
			"ifo",
			Option::ARG_REQUIRED,
			"IFO pattern to use for search. (Default: all)",
			"pattern" ) );

    retval.Add( Option( OPT_TYPE,
			"type",
			Option::ARG_REQUIRED,
			"Type pattern to use for search. (Default: all)",
			"pattern" ) );

    return retval;
  }

  MountPointStats::
  MountPointStats( CommandLineOptions& Args )
    : m_args( Args ),
      m_ifo( "all" ),
      m_type( "all" )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );

      while( parsing )
      {
	switch( m_args.Parse( m_options, arg_name, arg_value ) )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_IFO:
	  m_ifo = arg_value;
	  break;
	case OPT_TYPE:
	  m_type = arg_value;
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& MountPointStats::
  Options( )
  {
    return m_options;
  }

  void MountPointStats::
  operator()( )
  {
    static const char pattern[] = "} {";
    std::string
      results( getHashNumbers( m_ifo.c_str( ), m_type.c_str( ) ) );

    size_t	b = 0;
    size_t	e = results.find( pattern );
    while ( 1 )
    {
      std::cout << results.substr( b,
				   ( e == std::string::npos )
				   ? e
				   : ( ++e - b ) )
		<< std::endl
	;
      if ( e == std::string::npos )
      {
	break;
      }
      b = ++e;
      e = results.find( pattern, b );
    }
  }

  //=====================================================================
  // Scan
  //=====================================================================
  OptionSet& Scan::m_options( Scan::init_options( ) );

  OptionSet& Scan::
  init_options( )
  {
    static OptionSet	retval;

    retval.
      Synopsis( "Subcommand: scan" );

    retval.
      Summary( "The scan sub command is intended to"
	       " scan a set of directories for files of interest"
	       " and generate a memory cache image."
	       );

    retval.Add( Option( OPT_CONCURRENCY,
			"concurrency",
			Option::ARG_REQUIRED,
			"Number of mount points to scan concurrently",
			"integer" ) );

    retval.Add( Option( OPT_EXTENSIONS,
			"extensions",
			Option::ARG_REQUIRED,
			"Comma seperated list of file extensions",
			"list" ) );

    retval.Add( Option( OPT_MOUNT_POINTS,
			"mount-points",
			Option::ARG_REQUIRED,
			"Comma seperated list of mount points to scan",
			"list" ) );

    retval.Add( Option( OPT_OUTPUT,
			"output",
			Option::ARG_REQUIRED,
			"Filename for the output",
			"filename" ) );

    retval.Add( Option( OPT_TYPE,
			"type",
			Option::ARG_REQUIRED,
			"Type pattern to use for search. (Default: all)",
			"pattern" ) );

    retval.Add( Option( OPT_VERSION,
			"version",
			Option::ARG_REQUIRED,
			"Version of the binary diskcache dump format to output",
			"version" ) );
    return retval;
  }

  Scan::
  Scan( CommandLineOptions& Args )
    : m_args( Args ),
      m_mode( diskCache::CacheFile::CF_MODE_BINARY ),
      m_output( "" ),
      m_version( diskCache::CacheFile::CURRENT_VERSION )
  {
    if ( m_args.empty( ) == false )
    {
      //-----------------------------------------------------------------
      // Parse the commands
      //-----------------------------------------------------------------
      std::string	arg_name;
      std::string	arg_value;
      bool 		parsing( true );

      while( parsing )
      {
	switch( m_args.Parse( m_options, arg_name, arg_value ) )
	{
	case General::CommandLineOptions::OPT_END_OF_OPTIONS:
	  parsing = false;
	  break;
	case OPT_CONCURRENCY:
	  {
	    std::istringstream	value_stream( arg_value );

	    INT_4U	concurrency;

	    value_stream >> concurrency;

	    ScanConcurrency( concurrency );
	  }
	  break;
	case OPT_EXTENSIONS:
	  {
	    //-----------------------------------------------------------
	    // Generate list of extensions
	    //-----------------------------------------------------------
	    size_t pos = 0;
	    size_t end = 0;

	    while ( end != std::string::npos )
	    {
	      end = arg_value.find_first_of( ",", pos );
	      m_extensions.push_back( arg_value.substr( pos,( ( end == std::string::npos )
							      ? end
							      : end - pos ) ) );
	      pos = end + 1;
	    }
	  }
	  break;
	case OPT_MOUNT_POINTS:
	  {
	    //-----------------------------------------------------------
	    // Generate list of mount points
	    //-----------------------------------------------------------
	    size_t pos = 0;
	    size_t end = 0;

	    while ( end != std::string::npos )
	    {
	      end = arg_value.find_first_of( ",", pos );
	      m_mount_points.push_back( arg_value.substr( pos,( ( end == std::string::npos )
								? end
								: end - pos ) ) );
	      pos = end + 1;
	    }
	  }
	  break;
	case OPT_TYPE:
	  m_type = arg_value;
	  break;
	case OPT_OUTPUT:
	  m_output = arg_value;
	  break;
	case OPT_VERSION:
	  {
	    std::istringstream	version_stream( arg_value );

	    version_stream >> m_version;
	    diskCache::CacheFile::ValidateVersion( m_version );
	  }
	  break;
	default:
	  break;
	}
      }
    }
  }

  const OptionSet& Scan::
  Options( )
  {
    return m_options;
  }

  void Scan::
  operator()( )
  {
    bool		first( true );
    std::ostringstream	data;

    //-------------------------------------------------------------------
    // Add the extensions
    //-------------------------------------------------------------------
    first = true;
    data.str( "" );

    for ( extensions_type::const_iterator
	    cur = m_extensions.begin( ),
	    last = m_extensions.end( );
	  cur != last;
	  ++cur )
    {
      if ( first )
      {
	first = false;
      }
      else
      {
	data << " ";
      }
      data << *cur;
    }
    updateFileExtList( data.str( ).c_str( ) );

    //-------------------------------------------------------------------
    // Add the mount points
    //-------------------------------------------------------------------
    first = true;
    data.str( "" );

    for ( mount_points_type::const_iterator
	    cur = m_mount_points.begin( ),
	    last = m_mount_points.end( );
	  cur != last;
	  ++cur )
    {
      if ( first )
      {
	first = false;
      }
      else
      {
	data << " ";
      }
      data << *cur;
    }
    updateMountPtList( data.str( ).c_str( ), false );

    //-------------------------------------------------------------------
    // Scan the directories
    //-------------------------------------------------------------------

    ScanMountPointList( &std::cout );

    //-------------------------------------------------------------------
    // Dump the contents
    //-------------------------------------------------------------------
    if ( m_output.empty( ) )
    {
      std::string cache( getDirCache( ) );

      if ( cache.empty( ) == false )
      {
	size_t	b = 0;
	size_t	e = cache.find( " /" );
	while ( 1 )
	{
	  std::cout << cache.substr( b,
				     ( e == std::string::npos )
				     ? e
				     : ( e - b ) )
		    << std::endl
	    ;
	  if ( e == std::string::npos )
	  {
	    break;
	  }
	  b = e + 1;
	  e = cache.find( " /", b );
	}
      }
    }
    else
    {
      //-----------------------------------------------------------------
      /// \todo
      ///     When the output is binary, the query options should allow
      ///     the user to request a subset of the entire cache.
      //-----------------------------------------------------------------
      diskCache::CacheFile	cf;

      cf.writeToFile( m_output.c_str( ),
		      m_mode,
		      m_version );
    }
  }
}

//=======================================================================
//
//=======================================================================
void
Usage( const CommandLineOptions::option_type& ProgramName,
	const OptionSet& Opts )
{
  std::cout << "Usage: "
	    << ProgramName
	    << Opts
	    << std::endl
    ;
}

