// System Header Files

#include <sys/stat.h>

#include <list>
#include <memory>
#include <sstream>

// General Header Files   
#include "general/unittest.h"

#include "general/types.hh"
#include "general/Directory.hh"
#include "general/SharedPtr.hh"

#include "diskcache.hh"
#include "dirhash.hh"
#include "dirinfo.hh"
#include "mountpt.hh"

using General::Directory;
using diskCache::DirectoryHash;
using diskCache::DirectoryInfo;
using diskCache::LogMessage;
using diskCache::MountPointHash;

//---------------------------------------------------------------
//---------------------------------------------------------------
General::UnitTest	Test;

typedef std::pair< INT_4U, INT_4U > range_type;
typedef std::list< range_type > ranges_type;
typedef General::SharedPtr< DirectoryInfo > directory_info_ptr_type;

static void test_dir_cleanup( );
static const std::string& test_dir_name( );
static void test_dir_setup( INT_4U FrameDT,
			    const ranges_type& Ranges );
static void test_dir_populate( INT_4U FrameDT,
			       const ranges_type& Ranges );

static void test_multi_gap( );

//---------------------------------------------------------------
//---------------------------------------------------------------
int
main( int argc, char **argv )
{
  Test.Init( argc, argv );
  
  MountPointHash::updateMountPtList( test_dir_name( ).c_str( ),
				     false );

  MountPointHash::updateFileExtList( ".gwf" );

  //-------------------------------------------------------------
  // Perform tests
  //-------------------------------------------------------------
  test_multi_gap( );

  //-------------------------------------------------------------
  // Cleanup and exit
  //-------------------------------------------------------------
  // test_dir_cleanup( );
  Test.Exit();
   
  return 0;   
}

//---------------------------------------------------------------
// 
//---------------------------------------------------------------

static void
test_dir_cleanup( )
{
  const std::string&	dir_name( test_dir_name( ) );
  //-------------------------------------------------------------
  // Open the directory and remove all of its contents.
  //-------------------------------------------------------------
  try
  {
    Directory	dir( dir_name );

    if ( dir.Fd( ) >= 0 )
    {
      while( dir.Next( ) )
      {
	std::string	filename( dir_name );
	if ( ( strcmp( dir.EntryName( ), "." ) == 0 ) ||
	     ( strcmp( dir.EntryName( ), ".." ) == 0 ) )
	{
	  // Cannot remove current or parent directory entries.
	  continue;
	}
	filename += "/";
	filename += dir.EntryName( );
	unlink( filename.c_str( ) );
      }
      dir.Close( );
      rmdir( dir_name.c_str( ) );
    }
    else
    {
      unlink( dir_name.c_str( ) );
    }
  }
  catch( ... )
  {
  }
}

//---------------------------------------------------------------
// 
//---------------------------------------------------------------
static const std::string&
test_dir_name( )
{
  static std::string	dir;
  if ( dir.length( ) == 0 )
  {
    char cwd[2048];

    if ( getcwd( cwd, sizeof( cwd ) ) != (char*)NULL )
    {
      dir = cwd;
    }
    dir += "/pr3404.d";
  }
  return dir;
}

//---------------------------------------------------------------
// 
//---------------------------------------------------------------
static void
test_dir_populate( INT_4U FrameDT,
		const ranges_type& Ranges )
{
  //-------------------------------------------------------------
  // Insure there is nothing left from a previous run
  //-------------------------------------------------------------
  const std::string&	dir_name( test_dir_name( ) );


  for ( ranges_type::const_iterator
	  cur = Ranges.begin( ),
	  last = Ranges.end( );
	cur != last;
	++cur )
  {
    for ( INT_4U
	    start = cur->first;
	  start < cur->second;
	  start += FrameDT )
    {
      std::ostringstream	filename;

      filename << dir_name << "/"
	       << "Z-pr-" << start << "-" << FrameDT
	       << ".gwf";
      std::ofstream( filename.str( ).c_str( ) );
    }
  }
}


//---------------------------------------------------------------
// 
//---------------------------------------------------------------
static void
test_dir_setup( INT_4U FrameDT,
		const ranges_type& Ranges )
{
  //-------------------------------------------------------------
  // Insure there is nothing left from a previous run
  //-------------------------------------------------------------
  const std::string&	dir_name( test_dir_name( ) );

  test_dir_cleanup( );
  //-------------------------------------------------------------
  // Create the directory.
  //-------------------------------------------------------------
  mkdir( dir_name.c_str( ), 0755 );
  test_dir_populate( FrameDT, Ranges );
}


//---------------------------------------------------------------
// 
//---------------------------------------------------------------

static void
test_multi_gap( )
{
  const INT_4U dt( 10 );

  ranges_type	ranges;

  std::list< std::string >	edirs;

  //-------------------------------------------------------------
  // Create a gappy directory
  //-------------------------------------------------------------
  ranges.push_back( range_type( 100, 120 ) );
  ranges.push_back( range_type( 140, 160 ) );
  ranges.push_back( range_type( 180, 200 ) );
  ranges.push_back( range_type( 220, 240 ) );

  test_dir_setup( dt, ranges );

  //-------------------------------------------------------------
  // Scan the directory, checking to see the correct number
  // of files were discovered.
  //-------------------------------------------------------------


  {
    LogMessage			errors;

    DirectoryHash::Instance( ).scanDirectory( test_dir_name( ),
					      test_dir_name( ),
					      edirs,
					      errors,
					      -1 );

    std::cerr << "DEBUG: error: "
	      << errors( ).str( )
	      << std::endl
      ;
  }

#if 0 || 1
  //-------------------------------------------------------------
  // Fill in the gaps
  //-------------------------------------------------------------
  ranges.erase( ranges.begin( ),
		ranges.end( ) );

  ranges.push_back( range_type( 120, 140 ) );
  ranges.push_back( range_type( 160, 180 ) );
  ranges.push_back( range_type( 200, 220 ) );
  ranges.push_back( range_type( 240, 260 ) );
  ranges.push_back( range_type( 300, 320 ) );
  test_dir_populate( dt, ranges );
#endif /* 0 */


  //-------------------------------------------------------------
  // Scan the directory, checking to see the correct number
  // of files were discovered.
  //-------------------------------------------------------------

  {
    LogMessage			errors;

    DirectoryHash::Instance( ).scanDirectory( test_dir_name( ),
					      test_dir_name( ),
					      edirs,
					      errors,
					      -1 );

    std::cerr << "DEBUG: error: " << errors( ).str( )
	      << std::endl
      ;
  }

  //---------------------------------------------------------------------
  // Conflicting data
  //---------------------------------------------------------------------
  ranges.erase( ranges.begin( ),
		ranges.end( ) );

  ranges.push_back( range_type( 90, 90 + ( dt * 2 ) *40 ) );
  test_dir_populate( dt * 2, ranges );

  {
    LogMessage			errors;

    DirectoryHash::Instance( ).scanDirectory( test_dir_name( ),
					      test_dir_name( ),
					      edirs,
					      errors,
					      -1 );

    std::cerr << "DEBUG: error: " << errors( ).str( )
	      << std::endl
      ;
  }

  //---------------------------------------------------------------------
  // Second pass over conflicting data.
  //---------------------------------------------------------------------
  {
    //-------------------------------------------------------------------
    // Because data is in conflict, the directory should be re-scanned.
    //-------------------------------------------------------------------
    LogMessage			errors;

    DirectoryHash::Instance( ).scanDirectory( test_dir_name( ),
					      test_dir_name( ),
					      edirs,
					      errors,
					      -1 );

    std::cerr << "DEBUG: error: " << errors( ).str( )
	      << std::endl
      ;
  }

  // test_dir_cleanup( );
}
