#ifndef FrameCPP_VERSION_8_FrAdcData_HH
#define FrameCPP_VERSION_8_FrAdcData_HH

#include <memory>
#include <vector>

#include "general/types.hh"

#include "framecpp/Common/Container.hh"
#include "framecpp/Common/TOCInfo.hh"

#include "framecpp/Version7/FrAdcData.hh"

#include "framecpp/Version8/FrameSpec.hh"
#include "framecpp/Version8/Dimension.hh"
#include "framecpp/Version8/FrVect.hh"
#include "framecpp/Version8/STRING.hh"

namespace FrameCPP
{
  namespace Version_8
  {
    const REAL_4 FR_ADC_DATA_DEFAULT_BIAS = 0.0;
    const REAL_4 FR_ADC_DATA_DEFAULT_SLOPE = 1.0;
    const REAL_8 FR_ADC_DATA_DEFAULT_FSHIFT = 0.0;
    const REAL_8 FR_ADC_DATA_DEFAULT_TIME_OFFSET = 0.0;
    const INT_2U FR_ADC_DATA_DEFAULT_DATA_VALID = 0;
    const REAL_4 FR_ADC_DATA_DEFAULT_PHASE = 0.0;
    inline const char* FR_ADC_DATA_DEFAULT_UNITS( )
    {
      return "counts";
    }

    //===================================================================
    /// \brief ADC Data
    //===================================================================
    class FrAdcData
      : public ObjectWithChecksum< Common::FrameSpec::Object >,
	public Common::TOCInfo
    {
    public:
      typedef std::string name_type;
      typedef std::string comment_type;
      typedef INT_4U channelGroup_type;
      typedef INT_4U channelNumber_type;
      typedef INT_4U nBits_type;
      typedef REAL_4 bias_type;
      typedef REAL_4 slope_type;
      typedef std::string units_type;
      typedef REAL_8 sampleRate_type;
      typedef REAL_8 timeOffset_type;
      typedef REAL_8 fShift_type;
      typedef REAL_4 phase_type;
      typedef INT_2U dataValid_type;

      typedef Common::Container< FrVect >	data_type;
      typedef data_type::const_iterator const_iterator;
      typedef data_type::iterator iterator;

      typedef Common::Container< FrVect >	aux_type;
      typedef aux_type::const_iterator const_aux_iterator;
      typedef aux_type::iterator aux_iterator;

      static const REAL_4 DEFAULT_BIAS;
      static const REAL_4 DEFAULT_SLOPE;
      static const REAL_8 DEFAULT_FSHIFT;
      static const REAL_8 DEFAULT_TIME_OFFSET;
      static const INT_2U DEFAULT_DATA_VALID;
      static const REAL_4 DEFAULT_PHASE;

      inline static const char* DEFAULT_UNITS( )
      {
	return FR_ADC_DATA_DEFAULT_UNITS( );
      }

      //-----------------------------------------------------------------
      /// \brief Default constructor
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrAdcData( );

      //-----------------------------------------------------------------
      /// \brief Copy Constructor
      ///
      /// \param[in] Source
      ///     The object from which to copy the information.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrAdcData( const FrAdcData& Source );

      //-----------------------------------------------------------------
      /// \brief Constructor
      ///
      /// \param[in] name
      ///     The Channel name -- must be unique within the frame
      /// \param[in] group
      ///     Channel grouping number containing ADC.
      /// \param[in] nBits
      ///     Number of bits in A/D output.
      /// \param[in] bias
      ///     DC bias on channel (Units @ADC counts = 0)
      /// \param[in] slope
      ///     ADC calibration: input units/count.
      /// \param[in] units
      ///     ADC calibration: input units for slope.
      /// \param[in] fShift
      ///     The frequency (in Hz) in the origional data that
      ///     oorresponds to 0 Hz in the heterodyned series.
      /// \param[in] timeOffset
      ///     Offset of the first sample relative to the frame start
      ///     time (seconds). Must be positive and smaller than the
      ///     frame length.
      /// \param[in] dataValid
      ///     Data valid flag.
      /// \param[in] phase
      ///     Phase of heterodyning signal at start of dataset.
      ///
      /// \return
      ///     A new instance of the object.
      //-----------------------------------------------------------------
      FrAdcData( const std::string& name, INT_4U group,
		 INT_4U channel,
		 INT_4U nBits,
		 REAL_8 sampleRate,
		 REAL_4 bias = DEFAULT_BIAS,
		 REAL_4 slope = DEFAULT_SLOPE,
		 const std::string& units = DEFAULT_UNITS( ),
		 REAL_8 fShift = DEFAULT_FSHIFT,
		 REAL_8 timeOffset = DEFAULT_TIME_OFFSET,
		 INT_2U dataValid = DEFAULT_DATA_VALID,
		 REAL_4 phase = DEFAULT_PHASE );

      //-----------------------------------------------------------------
      /// \brief Promotion Constructor
      ///
      /// \param[in] Source
      ///     An FrAdcData structure from a previous frame specification.
      /// \param[in] Stream
      ///     The stream from which the earliest version of the
      ///     FrAdcData structure was read.
      ///
      /// \return
      ///     A new instance of the object.
      //-----------------------------------------------------------------
      FrAdcData( Previous::FrAdcData& Source, Common::IStream* Stream );

      //-----------------------------------------------------------------
      /// \brief Destructor
      //-----------------------------------------------------------------
      ~FrAdcData( );

      //-----------------------------------------------------------------
      /// \brief The name structure name of this object.
      ///
      /// \return
      ///     The name of the structure as specified by the frame
      ///     specification.
      //-----------------------------------------------------------------
      static const char* const StructName( );

      //-----------------------------------------------------------------
      /// \brief The description of structure
      ///
      /// \return
      ///     A Description object which describes this structure as
      ///     specified by the frame specification.
      //-----------------------------------------------------------------
      static const Common::Description* StructDescription( );

      //-----------------------------------------------------------------
      /// \brief equality operator
      ///
      /// \param[in] RHS
      ///     The FrAdcData object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      bool operator==( const FrAdcData& RHS ) const;

      //-----------------------------------------------------------------
      /// \brief inequality operator
      ///
      /// \param[in] RHS
      ///     The FrAdcData object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is not equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      bool operator!=( const FrAdcData& RHS ) const;

      //-----------------------------------------------------------------
      /// \brief equality operator for abstract data type
      ///
      /// \param[in] RHS
      ///     The object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      //-----------------------------------------------------------------
      /// \brief The name structure name of this object.
      ///
      /// \return
      ///     The name of the structure as specified by the frame
      ///     specification.
      //-----------------------------------------------------------------
      virtual const char* const ObjectStructName( ) const;

      //-----------------------------------------------------------------
      /// \brief Return the name associate with the FrAdcData structure.
      ///
      /// \return
      ///     The name associated with the FrAdcData structure
      //-----------------------------------------------------------------
      const std::string& GetName() const; 

      //-----------------------------------------------------------------
      /// \brief A brief description of the FrAdcData structure
      ///
      /// \return
      ///     A brief description of the FrAdcData structure
      //-----------------------------------------------------------------
      const std::string& GetComment( ) const;

      //-----------------------------------------------------------------
      /// \brief The channel number for the FrAdcData
      ///
      /// \return
      ///     The channel number for the FrAdcData
      //-----------------------------------------------------------------
      INT_4U GetChannelNumber( ) const;

      //-----------------------------------------------------------------
      /// \brief The channel group for the FrAdcData
      ///
      /// \return
      ///     The channel group for the FrAdcData
      //-----------------------------------------------------------------
      INT_4U GetChannelGroup( ) const;

      //-----------------------------------------------------------------
      /// \brief Tne number of bits in the A/D output.
      ///
      /// \return
      ///      Tne number of bits in the A/D output.
      //-----------------------------------------------------------------
      INT_4U GetNBits( ) const;

      //-----------------------------------------------------------------
      /// \brief DC bias on channel.
      ///
      /// \return
      ///     DC bias on channel.
      //-----------------------------------------------------------------
      REAL_4 GetBias( ) const;

      //-----------------------------------------------------------------
      /// \brief ADC calibration units per count.
      ///
      /// \return
      ///     ADC calibration units per count.
      //-----------------------------------------------------------------
      REAL_4 GetSlope( ) const;

      //-----------------------------------------------------------------
      /// \brief The units used in the ADC calibration
      ///
      /// \return
      ///     The units used in the ADC calibration
      //-----------------------------------------------------------------
      STRING GetUnits( ) const;

      //-----------------------------------------------------------------
      /// \brief The data acquisition rate.
      ///
      /// \return
      ///     The data acquisition rate.
      //-----------------------------------------------------------------
      REAL_8 GetSampleRate( ) const;

      //-----------------------------------------------------------------
      /// \brief The offset of the first sample relative to the frame start time.
      ///
      /// \return
      ///     The offset of the first sample relative to the frame
      ///     start time.
      //-----------------------------------------------------------------
      REAL_8 GetTimeOffset( ) const;

      //-----------------------------------------------------------------
      /// \brief The frequency in the original data corresponding to 0 Hz.
      ///
      /// \return
      ///     The frequency in the original data corresponding to 0 Hz.
      //-----------------------------------------------------------------
      REAL_8 GetFShift( ) const;

      //-----------------------------------------------------------------
      /// \brief The phase of hetrodyning signal at start of the dataset.
      ///
      /// \return
      ///     The phase of hetrodyning signal at start of the dataset.
      //-----------------------------------------------------------------
      REAL_4 GetPhase( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the dataValid value.
      ///
      /// \return
      ///     The value of the dataValid field.
      //-----------------------------------------------------------------
      INT_2U GetDataValid( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for vector for user-provided information.
      ///
      /// \return
      ///     A constant reference to the user-provided information.
      //-----------------------------------------------------------------
      const aux_type& RefAux( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for vector for user-provided information.
      ///
      /// \return
      ///     A reference to the user-provided information.
      //-----------------------------------------------------------------
      aux_type& RefAux( );

      //-----------------------------------------------------------------
      /// \brief Identifier for vector of sampled data.
      ///
      /// \return
      ///     A constant reference to the sampled data.
      //-----------------------------------------------------------------
      const data_type& RefData( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for vector of sampled data.
      ///
      /// \return
      ///     A reference to the sampled data.
      //-----------------------------------------------------------------
      data_type& RefData( );
   
      //-----------------------------------------------------------------
      /// \brief Extend the comment associated with the FrAdcData structure.
      //-----------------------------------------------------------------
      void AppendComment( const std::string& comment );

      //-----------------------------------------------------------------
      /// \brief Merge with another FrAdcData
      ///
      /// \param[in] RHS
      ///     The source of the information to append to this FrAdcData
      ///     structure.
      ///
      /// \return
      ///     A reference to this object
      //-----------------------------------------------------------------
      FrAdcData& Merge( const FrAdcData& RHS );

      //-----------------------------------------------------------------
      /// \brief Promote previous version of object to this version
      ///
      /// \param[in] Source
      ///     The frame specification version of the object.
      /// \param[in] Obj
      ///     The previous version of the object.
      /// \param[in] Stream
      ///     Stream from which the origional object was created.
      ///
      /// \return
      ///     A new instance of FrameH based on Obj
      //-----------------------------------------------------------------
      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      //-----------------------------------------------------------------
      /// \brief Establish the value of the dataValid field.
      ///
      /// \param[in] flag
      ///     The desired value of the dataValid field.
      //-----------------------------------------------------------------
      void SetDataValid( INT_2U flag );

      //-----------------------------------------------------------------
      /// \brief Establish the value of the bias field.
      ///
      /// \param[in] bias
      ///     The desired value of the bias field.
      //-----------------------------------------------------------------
      void SetBias( REAL_4 bias );

      //-----------------------------------------------------------------
      /// \brief Establish the value of the slope field.
      ///
      /// \param[in] slope
      ///     The desired value of the slope field.
      //-----------------------------------------------------------------
      void SetSlope( REAL_4 slope );

      //-----------------------------------------------------------------
      /// \brief Establish the value of the timeOffset field.
      ///
      /// \param[in] time
      ///     The desired value of the timeOffset field.
      //-----------------------------------------------------------------
      void SetTimeOffset( REAL_8 time );

      //-----------------------------------------------------------------
      /// \brief Request a subset of the data.
      ///
      /// \param[in] Offset
      ///     The number of seconds from the start of the FrAdcData
      ///     object from which to start extracting data.
      /// \param[in] Dt
      ///     The number of seconds of data in the resulting FrAdcData
      ///     strucutre.
      ///
      /// \return
      ///     A FrAdcData structure which contains Dt seconds of data
      ///     starting at Offset of this FrAdcData structure.
      //-----------------------------------------------------------------
      std::auto_ptr< FrAdcData >
      SubFrAdcData( REAL_8 Offset, REAL_8 Dt ) const;
   
      //-----------------------------------------------------------------
      /// \brief Answer requests for information as needed by TOC
      ///
      /// \param[in] InfoClass
      ///     Class of information being requested.
      //-----------------------------------------------------------------
      virtual void TOCQuery( int InfoClass, ... ) const;

    protected:
      //-----------------------------------------------------------------
      /// \brief Down grade an object
      ///
      /// \param[in] Target
      ///     The version of the frame specification to demote too.
      /// \param[in] Stream
      ///     The input stream from which the original object was read.
      ///
      /// \return
      ///     An object of the previous generation.
      //-----------------------------------------------------------------
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Upgrade an object
      ///
      /// \param[in] Target
      ///     The version of the frame specification to promote too.
      /// \param[in] Stream
      ///     The input stream from which the original object was read.
      ///
      /// \return
      ///     An object of this class.
      //-----------------------------------------------------------------
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Number of bytes needed to write this structure
      ///
      /// \param[in] Stream
      ///     The stream from which to the object is being read or
      ///     written.
      ///
      /// \return
      ///     The number of bytes need to read or write this object.
      //-----------------------------------------------------------------
      cmn_streamsize_type pBytes( const Common::StreamBase& Stream) const;

      //-----------------------------------------------------------------
      /// \brief Virtual constructor
      ///
      /// \param[in] Stream
      ///     The input stream from where the object is being read.
      //-----------------------------------------------------------------
      virtual FrAdcData* pCreate( Common::IStream& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Write the structure to the stream
      ///
      /// \param[in] Stream
      ///     The output stream where the object is to be written.
      //-----------------------------------------------------------------
      virtual void pWrite( Common::OStream& Stream ) const;

    private:
      //-----------------------------------------------------------------
      /// \brief Object with checksum data.
      //-----------------------------------------------------------------
      typedef ObjectWithChecksum< Common::FrameSpec::Object > object_type;

      using Common::FrameSpec::Object::Create;

      struct data_definition_type {
	//---------------------------------------------------------------
	/// Channel name  -- must be unique with the frame
	//---------------------------------------------------------------
	STRING	name;
	//---------------------------------------------------------------
	/// Comment for the ADC.
	//---------------------------------------------------------------
	STRING	comment;
	//---------------------------------------------------------------
	/// Channel grouping number containing ADC.
	///
	/// \note
	///     This variable is determined by the site and must be
	///     unique over all detectors.
	//---------------------------------------------------------------
	INT_4U	channelGroup;
	//---------------------------------------------------------------
	/// Channel number
	///
	/// \note
	///     This variable is determined by the site and must be
	///     unique over all detectors.
	//---------------------------------------------------------------
	INT_4U	channelNumber;
	//---------------------------------------------------------------
	/// Number of bits in A/D output.
	//---------------------------------------------------------------
	INT_4U	nBits;
	//---------------------------------------------------------------
	/// DC bias on channel (Units @ ADC counts = 0)
	//---------------------------------------------------------------
	REAL_4	bias;
	//---------------------------------------------------------------
	/// ADC calibration: input units/count.
	//---------------------------------------------------------------
	REAL_4	slope;
	//---------------------------------------------------------------
	/// ADC calibration: input units for slope
	/// If dimensionless, then units == <NONE>, in CAPITALS
	/// (without <...>).
	//---------------------------------------------------------------
	STRING	units;
	//---------------------------------------------------------------
	/// Data acquisition rate, samples per second.
	//---------------------------------------------------------------
	REAL_8	sampleRate;
	//---------------------------------------------------------------
	/// Offset of first sample relative to the frame start time
	/// in seconds.
	/// The value must be positive and smaller than the frame length.
	///
	/// \note
	///     Time offsets always added together (to obtain decimal
	///     time offset: time of the first sample equals
	///     GTime + timeOffset.
	///     All quantities are ALWAYS positive.
	//---------------------------------------------------------------
	REAL_8	timeOffset;
	//---------------------------------------------------------------
	/// The frequency (in HZ) in the original data that corresponds
	/// to zero Hz in the heterodyned series.
	///
	/// \note
	///     In the heterodyning process the real time series in
	///     multiplied by cos[2pi fShift(t-to) + phase] to get
	///     the real part and by -sin[2pi fShift(t-t(0)) + phase] to
	///     get the imaginary part of the resulting complex time
	///     series.
	///     The time origin t(0) is the beginning of the frame.
	//---------------------------------------------------------------
	REAL_8	fShift;
	//---------------------------------------------------------------
	/// Phase (in radian) of heterodyning signal at start of dataset.
	///
	/// \note
	///     In the heterodyning process the real time series in
	///     multiplied by cos[2pi fShift(t-to) + phase] to get
	///     the real part and by -sin[2pi fShift(t-t(0)) + phase] to
	///     get the imaginary part of the resulting complex time
	///     series.
	///     The time origin t(0) is the beginning of the frame.
	//---------------------------------------------------------------
	REAL_4	phase;
	//---------------------------------------------------------------
	/// Data valid flag: If the value is zero, the ADC data is valid.
	/// If the value is non-zero, the ADC data should be suspect
	/// or not valid.
	//---------------------------------------------------------------
	INT_2U	dataValid;
	//---------------------------------------------------------------
	/// Identifier for vector of sampled data.
	//---------------------------------------------------------------
	data_type	data;
	//---------------------------------------------------------------
	/// Identifier fro vector for user-provided information.
	/// Its use is generic.
	//---------------------------------------------------------------
	aux_type	aux;

	//---------------------------------------------------------------
	//---------------------------------------------------------------
	data_definition_type( );

	//---------------------------------------------------------------
	//---------------------------------------------------------------
	data_definition_type( const data_definition_type& Source );

	//---------------------------------------------------------------
	/// \brief Comparison operator
	//---------------------------------------------------------------
	bool operator==( const data_definition_type& RHS ) const;

	//---------------------------------------------------------------
	//---------------------------------------------------------------
	bool operator!=( const data_definition_type& RHS ) const;

	//---------------------------------------------------------------
	//---------------------------------------------------------------
	void copy_core( const data_definition_type& Source );
      } m_data;

      //-----------------------------------------------------------------
      /// \brief Stream Constructor
      ///
      /// \param[in] Stream
      ///     The stream from which the object is being read.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrAdcData( Common::IStream& Stream );
    };

    inline std::auto_ptr< Common::FrameSpec::Object > FrAdcData::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Object::PromoteObject< Previous::FrAdcData, FrAdcData >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* const FrAdcData::
    StructName( )
    {
      static const char* const name = "FrAdcData";
      return name;
    }

    inline const std::string& FrAdcData::
    GetComment( ) const
    {
      return m_data.comment;
    }

    inline INT_4U FrAdcData::
    GetChannelNumber( ) const
    { 
      return m_data.channelNumber;
    }

    inline INT_4U FrAdcData::
    GetChannelGroup( ) const
    {
      return m_data.channelGroup;
    }

    inline INT_4U FrAdcData::
    GetNBits( ) const
    {
      return m_data.nBits;
    }

    inline REAL_4 FrAdcData::
    GetBias( ) const
    {
      return m_data.bias;
    }

    inline REAL_4 FrAdcData::
    GetSlope( ) const 
    {
      return m_data.slope;
    }

    inline STRING FrAdcData::
    GetUnits( ) const
    {
      return m_data.units;
    }

    inline REAL_8 FrAdcData::
    GetSampleRate( ) const
    {
      return m_data.sampleRate;
    }

    inline REAL_8 FrAdcData::
    GetTimeOffset( ) const
    {
      return m_data.timeOffset;
    }

    inline REAL_8 FrAdcData::
    GetFShift( ) const
    {
      return m_data.fShift;
    }

    inline REAL_4 FrAdcData::
    GetPhase( ) const
    {
      return m_data.phase;
    }

    inline INT_2U FrAdcData::
    GetDataValid( ) const
    {
      return m_data.dataValid;
    }

    inline const FrAdcData::aux_type& FrAdcData::
    RefAux( ) const
    {
      return m_data.aux;
    }

    inline FrAdcData::aux_type& FrAdcData::
    RefAux( )
    {
      return m_data.aux;
    }

    inline const FrAdcData::data_type& FrAdcData::
    RefData( ) const
    {
      return m_data.data;
    }

    inline FrAdcData::data_type& FrAdcData::
    RefData( )
    {
      return m_data.data;
    }

    inline void FrAdcData::
    SetDataValid( INT_2U flag )
    {
      m_data.dataValid = flag;
    }

    inline void FrAdcData::
    SetBias( REAL_4 bias )
    {
      m_data.bias = bias;
    }

    inline void FrAdcData::
    SetSlope( REAL_4 slope )
    {
      m_data.slope = slope;
    }

    inline void FrAdcData::
    SetTimeOffset( REAL_8 time )
    {
      m_data.timeOffset = time;
    }

    inline bool FrAdcData::data_definition_type::
    operator!=( const data_definition_type& RHS ) const
    {
      return ! ( *this == RHS );
    }

    inline void FrAdcData::
    AppendComment( const std::string& comment )
    {
      FrameCPP::Common::AppendComment( m_data.comment, comment );
    } 

    inline bool FrAdcData::
    operator==( const FrAdcData& RHS ) const
    {
      return ( this->m_data == RHS.m_data );
    }

    inline bool FrAdcData::
    operator!=( const FrAdcData& RHS ) const
    {
      return ! ( *this == RHS );
    }
  } // namespace - Version_8
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_8_FrAdcData_HH */
